@@ -27,6 +27,7 @@
 #ifndef __STDC_LIMIT_MACROS
 #  define __STDC_LIMIT_MACROS 1
 #endif
+#define __STDC_WANT_IEC_60559_TYPES_EXT__
 
 #include "cbor.h"
 #include "cborinternal_p.h"
@@ -314,7 +315,10 @@ static CborError container_to_pretty(CborStreamFunction stream, void *out, CborV
 
     if (!recursionsLeft) {
         printRecursionLimit(stream, out);
-        return err;     /* do allow the dumping to continue */
+        while (!cbor_value_at_end(it) && !err) {
+            err = cbor_value_advance(it);
+        }
+        return err;     /* do allow the dumping to continue (if the advance was OK) */
     }
 
     while (!cbor_value_at_end(it) && !err) {
